
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '1.25.7',
    'branch_nick': 'release-1.25',
    'revision_id': '7c7137a28811ad8d7060088806e319a629f35215',
    'revno': 10570
}
